;;; Directory Local Variables            -*- no-byte-compile: t -*-
;;; For more information see (info "(emacs) Directory Variables")

((nil . ((indent-tabs-mode . t)
         (tab-width . 8)
         (fill-column . 70)
         (sentence-end-double-space . t)))
 (emacs-lisp-mode . ((indent-tabs-mode . nil)))
 (lisp-data-mode . ((indent-tabs-mode . nil)))
 (org-mode . ((indent-tabs-mode . nil)
              (org-adapt-indentation . nil)
              (org-src-content-indentation . 0)
              (org-footnote-auto-adjust . t)
              (org-footnote-auto-label . t)
              (org-footnote-define-inline . nil)
              (org-footnote-section . "Footnotes")
              (org-hide-emphasis-markers . nil)))
 ("testing/lisp" . ((emacs-lisp-mode . ((elisp-flymake-byte-compile-load-path "./" "../../lisp/"))))))



