;; Generated package description from org.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "org" "10.0.20260705.370" "Outline-based notes management and organizer" '((emacs "28.2")) :commit "13a25caeb2f71b2ddf19a8fae3316195302e8959" :authors '(("Carsten Dominik" . "carsten.dominik@gmail.com")) :maintainer '("Ihor Radchenko" . "yantar92@posteo.net") :keywords '("outlines" "hypermedia" "calendar" "text") :url "https://orgmode.org")
